@@ -1,18 +1,18 @@
-/*  Zcmp - decompress and compare two files byte by byte
-    Copyright (C) 2010-2016 Antonio Diaz Diaz.
+/* Zcmp - decompress and compare two files byte by byte
+   Copyright (C) 2010-2020 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
@@ -50,33 +50,32 @@ namespace {
 
 void show_help()
   {
-  std::printf( "Zcmp compares two files ('-' means standard input), and if they differ,\n"
-               "tells the first byte and line number where they differ. Bytes and lines\n"
-               "are numbered starting with 1. If any given file is compressed, its\n"
-               "decompressed content is used. Compressed files are decompressed on the\n"
-               "fly; no temporary files are created.\n"
-               "\nThe supported formats are bzip2, gzip, lzip and xz.\n"
+  std::printf( "zcmp compares two files and, if they differ, writes to standard output the\n"
+               "first byte and line number where they differ. Bytes and lines are numbered\n"
+               "starting with 1. A hyphen '-' used as a file argument means standard input.\n"
+               "If any file given is compressed, its decompressed content is used. Compressed\n"
+               "files are decompressed on the fly; no temporary files are created.\n"
+               "\nThe formats supported are bzip2, gzip, lzip, and xz.\n"
                "\nUsage: zcmp [options] file1 [file2]\n"
-               "\nCompares <file1> to <file2>. If <file2> is omitted zcmp tries the\n"
+               "\nzcmp compares file1 to file2. The standard input is used only if file1 or\n"
+               "file2 refers to standard input. If file2 is omitted zcmp tries the\n"
                "following:\n"
-               "\n  1. If <file1> is compressed, compares its decompressed contents with\n"
-               "  the corresponding uncompressed file (the name of <file1> with the\n"
+               "\n  - If file1 is compressed, compares its decompressed contents with\n"
+               "  the corresponding uncompressed file (the name of file1 with the\n"
                "  extension removed).\n"
-               "\n  2. If <file1> is uncompressed, compares it with the decompressed\n"
-               "  contents of <file1>.[lz|bz2|gz|xz] (the first one that is found).\n"
-               "\n  3. If no suitable file is found, compares <file1> with data read from\n"
-               "  standard input.\n"
+               "\n  - If file1 is uncompressed, compares it with the decompressed\n"
+               "  contents of file1.[lz|bz2|gz|xz] (the first one that is found).\n"
                "\nExit status is 0 if inputs are identical, 1 if different, 2 if trouble.\n"
                "\nOptions:\n"
                "  -h, --help                        display this help and exit\n"
                "  -V, --version                     output version information and exit\n"
                "  -b, --print-bytes                 print differing bytes\n"
-               "  -i, --ignore-initial=<n>[,<n2>]   ignore differences in the first <n> bytes\n"
+               "  -i, --ignore-initial=<n>[:<n2>]   ignore differences in the first <n> bytes\n"
                "  -l, --list                        list position, value of all differing bytes\n"
                "  -M, --format=<list>               process only the formats in <list>\n"
                "  -n, --bytes=<n>                   compare at most <n> bytes\n"
                "  -N, --no-rcfile                   don't read runtime configuration file\n"
-               "  -O, --force-format=[<f1>][,<f2>]  force given formats (bz2, gz, lz, xz)\n"
+               "  -O, --force-format=[<f1>][,<f2>]  force the formats given (bz2, gz, lz, xz)\n"
                "  -q, --quiet                       suppress all messages\n"
                "  -s, --silent                      (same as --quiet)\n"
                "  -v, --verbose                     verbose mode (same as --list)\n"
@@ -84,7 +83,7 @@ void show_help()
                "      --gz=<command>                set compressor and options for gzip format\n"
                "      --lz=<command>                set compressor and options for lzip format\n"
                "      --xz=<command>                set compressor and options for xz format\n"
-               "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
+               "\nNumbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
                "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n" );
   show_help_addr();
   }
@@ -94,8 +93,8 @@ long long getnum( const char * const ptr, const char ** const tailp = 0,
                   const long long llimit = 0,
                   const long long ulimit = LLONG_MAX )
   {
-  errno = 0;
   char * tail;
+  errno = 0;
   long long result = strtoll( ptr, &tail, 0 );
   if( tail == ptr )
     {
@@ -106,9 +105,13 @@ long long getnum( const char * const ptr, const char ** const tailp = 0,
 
   if( !errno && tail[0] && std::isalpha( tail[0] ) )
     {
-    const int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
-    int exponent = 0;				// 0 = bad multiplier
-    switch( tail[0] )
+    const unsigned char ch = *tail++;
+    int factor;
+    bool bsuf;					// 'B' suffix is present
+    if( tail[0] == 'i' ) { ++tail; factor = 1024; } else factor = 1000;
+    if( tail[0] == 'B' ) { ++tail; bsuf = true; } else bsuf = false;
+    int exponent = -1;				// -1 = bad multiplier
+    switch( ch )
       {
       case 'Y': exponent = 8; break;
       case 'Z': exponent = 7; break;
@@ -119,8 +122,9 @@ long long getnum( const char * const ptr, const char ** const tailp = 0,
       case 'M': exponent = 2; break;
       case 'K': if( factor == 1024 ) exponent = 1; break;
       case 'k': if( factor == 1000 ) exponent = 1; break;
+      case 'B': if( factor == 1000 && !bsuf ) exponent = 0; break;
       }
-    if( exponent <= 0 )
+    if( exponent < 0 )
       {
       show_error( "Bad multiplier in numerical argument.", 0, true );
       std::exit( 2 );
@@ -146,9 +150,14 @@ void parse_ignore_initial( const char * const arg, long long ignore_initial[2] )
   {
   const char * tail;
   ignore_initial[0] = getnum( arg, &tail );
-  if( *tail == ',' || *tail == ':' )
+  if( *tail == ':' || *tail == ',' )
     ignore_initial[1] = getnum( ++tail );
-  else ignore_initial[1] = ignore_initial[0];
+  else if( *tail == 0 ) ignore_initial[1] = ignore_initial[0];
+  else
+    {
+    show_error( "Bad separator in argument of '--ignore-initial'", 0, true );
+    std::exit( 2 );
+    }
   }
 
 
@@ -233,7 +242,7 @@ int cmp( const long long max_size, const int infd[2],
       rd[i] = readblock( infd[i], buffer[i], size );
       if( rd[i] != size && errno )
         {
-        show_error2( "Error reading file", filenames[i].c_str() );
+        show_file_error( filenames[i].c_str(), "Read error", errno );
         return 2;
         }
       }
@@ -264,6 +273,7 @@ int cmp( const long long max_size, const int infd[2],
                        filenames[0].c_str(), filenames[1].c_str(),
                        byte_number, line_number, c0, buf0, c1, buf1 );
           }
+        std::fflush( stdout );
         return 1;
         }
       else			// verbosity > 0 ; show all differences
@@ -286,6 +296,7 @@ int cmp( const long long max_size, const int infd[2],
               }
             }
           }
+        std::fflush( stdout );
         }
       }
 
@@ -314,8 +325,8 @@ int main( const int argc, const char * const argv[] )
   long long max_size = -1;			// < 0 means unlimited size
   int format_types[2] = { -1, -1 };
   bool print_bytes = false;
-  invocation_name = argv[0];
   program_name = "zcmp";
+  invocation_name = ( argc > 0 ) ? argv[0] : program_name;
 
   const Arg_parser::Option options[] =
     {
@@ -391,41 +402,47 @@ int main( const int argc, const char * const argv[] )
     STDIN_FILENO : open_instream( filenames[0] );
   if( infd[0] < 0 ) return 2;
 
-  if( ( files == 1 && filenames[0] == "-" ) ||
-      ( files == 2 && check_identical( filenames[0].c_str(),
-                                       filenames[1].c_str() ) ) )
-    {
-    if( ignore_initial[0] == ignore_initial[1] ) return 0;
-    else { show_error( "Can't compare parts of same file." ); return 2; }
-    }
-
   if( files == 2 )
     {
+    if( check_identical( filenames[0].c_str(), filenames[1].c_str() ) )
+      {
+      if( ignore_initial[0] == ignore_initial[1] ) return 0;
+      else { show_error( "Can't compare parts of same file." ); return 2; }
+      }
     infd[1] = ( filenames[1] == "-" ) ?
       STDIN_FILENO : open_instream( filenames[1] );
     if( infd[1] < 0 ) return 2;
     }
   else
     {
+    if( filenames[0] == "-" )
+      { show_error( "Missing operand after '-'.", 0, true ); return 2; }
     if( format_types[0] >= 0 || format_types[1] >= 0 )
       { show_error( "Two files must be given when format is specified.", 0, true );
         return 2; }
     filenames[1] = filenames[0];
     infd[1] = open_other_instream( filenames[1] );
-    if( infd[1] < 0 ) { infd[1] = STDIN_FILENO; filenames[1] = "-"; }
+    if( infd[1] < 0 )
+      {
+      if( verbosity >= 0 )
+        std::fprintf( stderr, "%s: Can't find file to compare with '%s'.\n",
+                      program_name, filenames[0].c_str() );
+      show_error( 0, 0, true ); return 2;
+      }
     }
 
   int old_infd[2];		// copy of file descriptors of the two files
   old_infd[0] = infd[0]; old_infd[1] = infd[1];
   Children children[2];
-  if( !set_data_feeder( &infd[0], children[0], format_types[0] ) ||
-      !set_data_feeder( &infd[1], children[1], format_types[1] ) )
+  if( !set_data_feeder( filenames[0], &infd[0], children[0], format_types[0] ) ||
+      !set_data_feeder( filenames[1], &infd[1], children[1], format_types[1] ) )
     return 2;
 
   for( int i = 0; i < 2; ++i )
     if( !skip_ignore_initial( ignore_initial[i], infd[i] ) )
       {
-      show_error2( "Can't skip initial bytes from file", filenames[i].c_str() );
+      show_file_error( filenames[i].c_str(),
+                       "Read error skipping initial bytes", errno );
       return 2;
       }
 
@@ -440,13 +457,13 @@ int main( const int argc, const char * const argv[] )
       { show_close_error(); retval = 2; }
     if( filenames[i] != "-" && close( old_infd[i] ) != 0 )
       {
-      show_error2( "Can't close input file", filenames[i].c_str() );
+      show_file_error( filenames[i].c_str(), "Error closing input file", errno );
       retval = 2;
       }
     }
   if( std::fclose( stdout ) != 0 )
     {
-    show_error( "Can't close stdout", errno );
+    show_error( "Error closing stdout", errno );
     retval = 2;
     }
 
