@@ -1,23 +1,24 @@
-/*  Zutils - Utilities dealing with compressed files
-    Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014 Antonio Diaz Diaz.
+/* Zutils - Utilities dealing with compressed files
+   Copyright (C) 2009-2020 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
 
 #include <cerrno>
+#include <csignal>
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
@@ -33,24 +34,21 @@
 
 namespace {
 
-// first magic byte must be different among formats
-enum { bzip2_magic_size = 3,
-        gzip_magic_size = 2,
-        lzip_magic_size = 4,
-          xz_magic_size = 5 };
-const uint8_t bzip2_magic[bzip2_magic_size] =
-  { 0x42, 0x5A, 0x68 };				// "BZh"
-const uint8_t gzip_magic[gzip_magic_size] =
-  { 0x1F, 0x8B };
-const uint8_t lzip_magic[lzip_magic_size] =
-  { 0x4C, 0x5A, 0x49, 0x50 };			// "LZIP"
-const uint8_t xz_magic[xz_magic_size] =
-  { 0xFD, 0x37, 0x7A, 0x58, 0x5A };		// 0xFD, "7zXZ"
+inline bool isvalid_ds( const uint8_t ds )	// lzip valid dictionary_size
+  {
+  enum { min_dictionary_size = 1 << 12,
+         max_dictionary_size = 1 << 29 };
+  unsigned dictionary_size = ( 1 << ( ds & 0x1F ) );
+  if( dictionary_size > min_dictionary_size )
+    dictionary_size -= ( dictionary_size / 16 ) * ( ( ds >> 5 ) & 7 );
+  return ( dictionary_size >= min_dictionary_size &&
+           dictionary_size <= max_dictionary_size );
+  }
 
 
-// Returns -1 if child not terminated, 2 in case of error, or
-// exit status of child process 'pid'.
-//
+/* Returns -1 if child not terminated, 2 in case of error, or exit status of
+   child process 'pid'.
+*/
 int child_status( const pid_t pid, const char * const name )
   {
   int status;
@@ -60,7 +58,7 @@ int child_status( const pid_t pid, const char * const name )
     if( tmp == -1 && errno != EINTR )
       {
       if( verbosity >= 0 )
-        std::fprintf( stderr, "%s: Error checking status of '%s': %s.\n",
+        std::fprintf( stderr, "%s: Error checking status of '%s': %s\n",
                       program_name, name, std::strerror( errno ) );
       _exit( 2 );
       }
@@ -68,60 +66,53 @@ int child_status( const pid_t pid, const char * const name )
     if( tmp == pid ) break;			// child terminated
     }
   if( WIFEXITED( status ) ) return WEXITSTATUS( status );
+  if( WIFSIGNALED( status ) && WTERMSIG( status ) == SIGPIPE ) return 0;
   return 2;
   }
 
 } // end namespace
 
 
-int parse_format_type( const std::string & arg )
-  {
-  for( int i = 0; i < num_formats; ++i )
-    if( arg == format_names[i] )
-      return i;
-  show_error( "Bad argument for '--format' option." );
-  std::exit( 1 );
-  }
-
-
-// Returns the number of bytes really read.
-// If (returned value < size) and (errno == 0), means EOF was reached.
-//
+/* Returns the number of bytes really read.
+   If (returned value < size) and (errno == 0), means EOF was reached.
+*/
 int readblock( const int fd, uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = read( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
+    const int n = read( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
     else if( n == 0 ) break;				// EOF
     else if( errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
-// Returns the number of bytes really written.
-// If (returned value < size), it is always an error.
-//
+/* Returns the number of bytes really written.
+   If (returned value < size), it is always an error.
+*/
 int writeblock( const int fd, const uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = write( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
+    const int n = write( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
     else if( n < 0 && errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
-bool feed_data( const int infd, const int outfd,
+// Empty filename means stdin.
+//
+bool feed_data( const std::string & filename, const int infd, const int outfd,
                 const uint8_t * magic_data, const int magic_size )
   {
   if( magic_size && writeblock( outfd, magic_data, magic_size ) != magic_size )
@@ -132,7 +123,8 @@ bool feed_data( const int infd, const int outfd,
     {
     const int size = readblock( infd, buffer, buffer_size );
     if( size != buffer_size && errno )
-      { show_error( "Read error", errno ); return false; }
+      { const char * const name = filename.empty() ? "-" : filename.c_str();
+        show_file_error( name, "Read error", errno ); return false; }
     if( size > 0 && writeblock( outfd, buffer, size ) != size )
       { show_error( "Write error", errno ); return false; }
     if( size < buffer_size ) break;
@@ -149,28 +141,31 @@ bool good_status( const Children & children, const bool finished )
     const pid_t pid = children.pid[i];
     if( pid )
       {
-      const char * const msg =
-        ( i & 1 ) ? children.compressor_name : "data feeder";
-      if( !finished )
+      const char * const name =
+        ( i == 0 ) ? "data feeder" : children.compressor_name;
+      // even if compressor finished, trailing data may remain in data feeder
+      if( i == 0 || !finished )
         {
-        const int tmp = child_status( pid, msg );
-        if( tmp < 0 ) kill( pid, SIGTERM );	// child not terminated
+        const int tmp = child_status( pid, name );
+        if( tmp < 0 )				// child not terminated
+          { kill( pid, SIGTERM ); wait_for_child( pid, name ); }
         else if( tmp != 0 ) error = true;	// child status != 0
         }
       else
-        if( wait_for_child( pid, msg ) != 0 ) error = true;
+        if( wait_for_child( pid, name ) != 0 ) error = true;
       }
     }
   return !error;
   }
 
 
-bool set_data_feeder( int * const infdp, Children & children, int format_index )
+bool set_data_feeder( const std::string & filename, int * const infdp,
+                      Children & children, int format_index )
   {
-  const uint8_t * magic_data = 0;
+  uint8_t magic_data[magic_buf_size];
   int magic_size = 0;
   if( format_index < 0 )
-    format_index = test_format( *infdp, &magic_data, &magic_size );
+    format_index = test_format( *infdp, magic_data, &magic_size );
   children.compressor_name = get_compressor_name( format_index );
 
   if( children.compressor_name )	// compressed
@@ -186,7 +181,7 @@ bool set_data_feeder( int * const infdp, Children & children, int format_index )
       {
       if( close( fda[0] ) != 0 ||
           close( fda2[0] ) != 0 || close( fda2[1] ) != 0 ||
-          !feed_data( old_infd, fda[1], magic_data, magic_size ) )
+          !feed_data( filename, old_infd, fda[1], magic_data, magic_size ) )
         _exit( 2 );
       if( close( fda[1] ) != 0 )
         { show_close_error(); _exit( 2 ); }
@@ -235,7 +230,7 @@ bool set_data_feeder( int * const infdp, Children & children, int format_index )
     if( pid == 0 )			// child (feeder)
       {
       if( close( fda[0] ) != 0 ||
-          !feed_data( old_infd, fda[1], magic_data, magic_size ) )
+          !feed_data( filename, old_infd, fda[1], magic_data, magic_size ) )
         _exit( 2 );
       if( close( fda[1] ) != 0 )
         { show_close_error(); _exit( 2 ); }
@@ -251,46 +246,38 @@ bool set_data_feeder( int * const infdp, Children & children, int format_index )
   }
 
 
-int test_format( const int infd, const uint8_t ** const magic_datap,
+// Returns format index or -1 if uncompressed
+//
+int test_format( const int infd, uint8_t magic_data[],
                  int * const magic_sizep )
   {
-  enum { buf_size = 5 };
-  static uint8_t buf[buf_size];
-  int i = 0;
-  if( readblock( infd, buf, 1 ) == 1 )
+  enum { bzip2_magic_size = 3,
+          gzip_magic_size = 2,
+          lzip_magic_size = 5,
+            xz_magic_size = 5 };
+  const uint8_t bzip2_magic[bzip2_magic_size] =
+    { 0x42, 0x5A, 0x68 };				// "BZh"
+  const uint8_t gzip_magic[gzip_magic_size] =
+    { 0x1F, 0x8B };
+  const uint8_t lzip_magic[lzip_magic_size] =
+    { 0x4C, 0x5A, 0x49, 0x50, 0x01 };			// "LZIP\001"
+  const uint8_t xz_magic[xz_magic_size] =
+    { 0xFD, 0x37, 0x7A, 0x58, 0x5A };			// 0xFD, "7zXZ"
+
+  *magic_sizep = readblock( infd, magic_data, magic_buf_size );
+  if( *magic_sizep == magic_buf_size )
     {
-    ++i;
-    if( buf[0] == bzip2_magic[0] )
-      {
-      if( readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == bzip2_magic[1] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == bzip2_magic[2] )
-        { *magic_datap = bzip2_magic; *magic_sizep = bzip2_magic_size;
-          return fmt_bz2; }
-      }
-    else if( buf[0] == gzip_magic[0] )
-      {
-      if( readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == gzip_magic[1] )
-        { *magic_datap = gzip_magic; *magic_sizep = gzip_magic_size;
-          return fmt_gz; }
-      }
-    else if( buf[0] == lzip_magic[0] )
-      {
-      if( readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == lzip_magic[1] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == lzip_magic[2] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == lzip_magic[3] )
-        { *magic_datap = lzip_magic; *magic_sizep = lzip_magic_size;
-          return fmt_lz; }
-      }
-    else if( buf[0] == xz_magic[0] )
-      {
-      if( readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == xz_magic[1] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == xz_magic[2] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == xz_magic[3] &&
-          readblock( infd, &buf[i], 1 ) == 1 && buf[i++] == xz_magic[4] )
-        { *magic_datap = xz_magic; *magic_sizep = xz_magic_size;
-          return fmt_xz; }
-      }
+    if( std::memcmp( magic_data, bzip2_magic, bzip2_magic_size ) == 0 &&
+        magic_data[3] >= '1' && magic_data[3] <= '9' &&
+        std::memcmp( magic_data + 4, "1AY&SY", 6 ) == 0 )
+      return fmt_bz2;
+    if( std::memcmp( magic_data, gzip_magic, gzip_magic_size ) == 0 )
+      return fmt_gz;
+    if( std::memcmp( magic_data, lzip_magic, lzip_magic_size ) == 0 &&
+        isvalid_ds( magic_data[lzip_magic_size] ) )
+      return fmt_lz;
+    if( std::memcmp( magic_data, xz_magic, xz_magic_size ) == 0 )
+      return fmt_xz;
     }
-  *magic_datap = buf; *magic_sizep = i;
   return -1;
   }
